// vim: set ft=cpp:
// vim: set foldmethod=marker:
#ifndef _FOURMOMENTUM_H_
#define _FOURMOMENTUM_H_

#include <iostream>
#include <string>
#include <complex>
#include <cmath>
#include <cstdlib>
#include <cfloat>
#include "mytypes.h"

template<typename T>
class FourMomentum {
  
 private:
  static DOUBLE tiny;
  T mom[4];
 public:

  // class constructors

   
  FourMomentum<T>() {
    mom[0] = mom[1] = mom[2] = mom[3] = 0.;
  }
      
  FourMomentum<T>(const std::string & s, const T & p0,
		  const T & p1, const T & p2, 
		  const T & p3) {
    mom[0] = p0;
    mom[1] = p1;
    mom[2] = p2;
    mom[3] = p3;
  }
  
  FourMomentum<T>(const T & p0,const T & p1, 
		  const T & p2, const T & p3){
    mom[0] = p0;
    mom[1] = p1;
    mom[2] = p2;
    mom[3] = p3;
  }
  
  template<typename F>
    FourMomentum<T>(const FourMomentum<F> &  p  ){
    
    mom[0] = p.getComponent(0);
    mom[1] = p.getComponent(1);
    mom[2] = p.getComponent(2);
    mom[3] = p.getComponent(3);
  }

  FourMomentum<T>(const T p[4]){
    
    mom[0] = p[0];
    mom[1] = p[1];
    mom[2] = p[2];
    mom[3] = p[3];
  }

  //
  // member functions
  //

  T getComponent(int i) const {
    return(mom[i]);
  }

  //
  // member functions for kinematics
  //


  void setFourMomentum(const FourMomentum<T> & p);
  void setFourMomentum(const T p[4]);
  void setFourMomentum(const T p0, const T p1, 
		       const T p2, const T p3);
  
  //
  // overloaded operators
  //  

  template<typename TT>
    friend std::ostream& operator<<(std::ostream& os,
				    const FourMomentum<TT> & p); 
  template<typename TT>
    friend const FourMomentum<TT> operator-(const FourMomentum<TT> & p1);

  template<typename TT>
    friend const FourMomentum<TT> & operator+(const FourMomentum<TT> & p1);

  template<typename TT>
    friend const FourMomentum<TT> operator-(const FourMomentum<TT> & p1, 
					    const FourMomentum<TT> & p2);

  template<typename TT>
    friend const FourMomentum<TT> operator+(const FourMomentum<TT> & p1, 
					    const FourMomentum<TT> & p2);

  template<typename TT, typename CC>
    friend FourMomentum<std::complex<DOUBLE> > 
    operator-(const FourMomentum<TT> & p1, const FourMomentum<CC> & p2);


  template<typename TT, typename CC>
    friend FourMomentum<std::complex<DOUBLE> > 
    operator+(const FourMomentum<TT> & p1, const FourMomentum<CC> & p2);


  template<typename TT>
    friend const FourMomentum<TT> operator*(const FourMomentum<TT> & p,
					    const DOUBLE & s );

  template<typename TT>
    friend const FourMomentum<TT> operator*(const DOUBLE & s,
					    const FourMomentum<TT> & p);

  template<typename TT>
    friend const FourMomentum<std::complex<DOUBLE> > 
    operator*(const std::complex<DOUBLE> & s, const FourMomentum<TT> & p);
  

  template<typename TT>
    friend const FourMomentum<std::complex<DOUBLE> >
    operator*(const FourMomentum<TT> & p, const std::complex<DOUBLE> & s);
  
  template<typename TT>
  friend const FourMomentum<TT> operator/(const FourMomentum<TT> & p2,
					  const DOUBLE & s1 );

  template<typename TT>
    friend FourMomentum<std::complex<DOUBLE> >
    operator/(const FourMomentum<TT> & p2,
	      const std::complex<DOUBLE> & s1 );

  
  //
  // friend functions
  //

  //
  // additional functions for spinor products
  //

  template<typename TT>
    friend TT dot(const FourMomentum<TT> & p1, const FourMomentum<TT> & p2);
  
  template<typename TT, typename CC>
    friend std::complex<DOUBLE> dot(const FourMomentum<TT> & p1,
				    const FourMomentum<CC> & p2);
  // center of mass energy
  template<typename TT>
    friend TT S(const FourMomentum<TT> & p1);

  // inverse S
  template<typename TT>
  friend TT IS(const FourMomentum<TT> & P);

  template<typename TT>
  friend std::complex<DOUBLE> pp(const FourMomentum<TT> & p);

  template<typename TT>
  friend std::complex<DOUBLE> pm(const FourMomentum<TT> & p);

  template<typename TT>
  friend std::complex<DOUBLE> pT(const FourMomentum<TT> & p);

  template<typename TT>
  friend std::complex<DOUBLE> pTb(const FourMomentum<TT> & p);

  // normalisation of spinor products
  template<typename TT>
  friend std::complex<DOUBLE> SpNorm(const FourMomentum<TT> & p);

  template<typename TT, typename UU>
  friend std::complex<DOUBLE> A(const FourMomentum<TT> & p1,
				const FourMomentum<UU> & p2);

  template<typename TT, typename UU>
  friend std::complex<DOUBLE> IA(const FourMomentum<TT> & p1,
				 const FourMomentum<UU> & p2);

  template<typename TT, typename UU>
  friend std::complex<DOUBLE> B(const FourMomentum<TT> & p1,
				const FourMomentum<UU> & p2);

  template<typename TT, typename UU>
  friend std::complex<DOUBLE> IB(const FourMomentum<TT> & p1,
				 const FourMomentum<UU> & p2);

  template<typename TT, typename UU,typename VV>
  friend std::complex<DOUBLE> AB(const FourMomentum<TT> & p1,
				 const FourMomentum<UU> & p2,
				 const FourMomentum<VV> & p3);

  template<typename TT, typename UU,typename VV>
  friend std::complex<DOUBLE> IAB(const FourMomentum<TT> & p1,
				  const FourMomentum<UU> & p2,
				  const FourMomentum<VV> & p3);

  template<typename TT, typename UU, typename VV>
  friend std::complex<DOUBLE> BA(const FourMomentum<TT> & p1,
				 const FourMomentum<UU> & p2,
				 const FourMomentum<VV> & p3);

  template<typename TT, typename UU, typename VV>
  friend std::complex<DOUBLE> IBA(const FourMomentum<TT> & p1,
				  const FourMomentum<UU> & p2,
				  const FourMomentum<VV> & p3);


  template<typename TT, typename UU, typename VV, typename WW>
  friend std::complex<DOUBLE> AA(const FourMomentum<TT> & p1,
				 const FourMomentum<UU> & p2,
				 const FourMomentum<VV> & p3,
				 const FourMomentum<WW> & p4);

  template<typename TT, typename UU, typename VV, typename WW>
  friend std::complex<DOUBLE> IAA(const FourMomentum<TT> & p1,
				  const FourMomentum<UU> & p2,
				  const FourMomentum<VV> & p3,
				  const FourMomentum<WW> & p4);

  template<typename TT, typename UU, typename VV, typename WW, typename XX>
  friend std::complex<DOUBLE> BB(const FourMomentum<TT> & p1,
				 const FourMomentum<TT> & p2,
				 const FourMomentum<TT> & p3, 
				 const FourMomentum<TT> & p4);

  template<typename TT, typename UU, typename VV, typename WW, typename XX>
  friend std::complex<DOUBLE> IBB(const FourMomentum<TT> & p1,
				  const FourMomentum<VV> & p2,
				  const FourMomentum<WW> & p3,
				  const FourMomentum<XX> & p4);

  template<typename TT, typename UU, typename VV, typename WW, typename XX>
  friend std::complex<DOUBLE> AB3(const FourMomentum<TT> & p1, 
				  const FourMomentum<UU> & p2,
				  const FourMomentum<VV> & p3, 
				  const FourMomentum<WW> & p4,
				  const FourMomentum<XX> & p5);

  template<typename TT, typename UU, typename VV, typename WW, typename XX>
  friend std::complex<DOUBLE> IAB3(const FourMomentum<TT> & p1,
				   const FourMomentum<UU> & p2,
				   const FourMomentum<VV> & p3, 
				   const FourMomentum<WW> & p4,
				   const FourMomentum<XX> & p5);

  template<typename TT, typename UU, typename VV, typename WW, typename XX>
  friend std::complex<DOUBLE> BA3(const FourMomentum<TT> & p1, 
				  const FourMomentum<UU> & p2,
				  const FourMomentum<VV> & p3, 
				  const FourMomentum<WW> & p4,
				  const FourMomentum<XX> & p5);

  template<typename TT, typename UU, typename VV, typename WW, typename XX,
           typename YY>
  friend std::complex<DOUBLE> AA4(const FourMomentum<TT> & p1, 
				  const FourMomentum<UU> & p2,
				  const FourMomentum<VV> & p3, 
				  const FourMomentum<WW> & p4,
				  const FourMomentum<XX> & p5, 
				  const FourMomentum<YY> & p6);

  template<typename TT, typename UU, typename VV, typename WW, typename XX,
           typename YY>
  friend std::complex<DOUBLE> IAA4(const FourMomentum<TT> & p1, 
				   const FourMomentum<UU> & p2,
				   const FourMomentum<VV> & p3, 
				   const FourMomentum<WW> & p4,
				   const FourMomentum<XX> & p5, 
				   const FourMomentum<YY> & p6);

  template<typename TT, typename UU, typename VV, typename WW, typename XX,
           typename YY>
  friend std::complex<DOUBLE> BB4(const FourMomentum<TT> & p1,
				  const FourMomentum<UU> & p2,
				  const FourMomentum<VV> & p3,
				  const FourMomentum<WW> & p4,
				  const FourMomentum<XX> & p5,
				  const FourMomentum<YY> & p6);

  template<typename TT, typename UU, typename VV, typename WW, typename XX,
           typename YY>
  friend std::complex<DOUBLE> IBB4(const FourMomentum<TT> & p1, 
				   const FourMomentum<UU> & p2,
				   const FourMomentum<VV> & p3, 
				   const FourMomentum<WW> & p4,
				   const FourMomentum<XX> & p5,
				   const FourMomentum<YY> & p6);

  template<typename TT, typename UU, typename VV, typename WW, typename XX,
    	typename YY,typename ZZ>
  friend std::complex<DOUBLE> AB5(const FourMomentum<TT> & p1, 
				  const FourMomentum<UU> & p2,
				  const FourMomentum<VV> & p3, 
				  const FourMomentum<WW> & p4,
				  const FourMomentum<XX> & p5,
				  const FourMomentum<YY> & p6,
				  const FourMomentum<YY> & p7);

  template<typename TT, typename UU, typename VV, typename WW, typename XX,
    	typename YY,typename ZZ>
  friend std::complex<DOUBLE> IAB5(const FourMomentum<TT> & p1, 
				  const FourMomentum<UU> & p2,
				  const FourMomentum<VV> & p3, 
				  const FourMomentum<WW> & p4,
				  const FourMomentum<XX> & p5,
				  const FourMomentum<YY> & p6,
				  const FourMomentum<YY> & p7);


  template<typename TT, typename UU, typename VV, typename WW, typename XX,
       typename YY,typename ZZ,typename TU>
  friend std::complex<DOUBLE> AA6(const FourMomentum<TT> & p1, 
				  const FourMomentum<UU> & p2,
				  const FourMomentum<VV> & p3, 
				  const FourMomentum<WW> & p4,
				  const FourMomentum<XX> & p5, 
				  const FourMomentum<YY> & p6,
				  const FourMomentum<ZZ> & p7,
				  const FourMomentum<TU> & p8);

  template<typename TT, typename UU, typename VV, typename WW, typename XX,
       typename YY,typename ZZ,typename TU>
  friend std::complex<DOUBLE> IAA6(const FourMomentum<TT> & p1, 
				  const FourMomentum<UU> & p2,
				  const FourMomentum<VV> & p3, 
				  const FourMomentum<WW> & p4,
				  const FourMomentum<XX> & p5, 
				  const FourMomentum<YY> & p6,
				  const FourMomentum<ZZ> & p7,
				  const FourMomentum<TU> & p8);

  template<typename TT, typename UU, typename VV, typename WW, typename XX,
       typename YY,typename ZZ,typename TU>
  friend std::complex<DOUBLE> BB6(const FourMomentum<TT> & p1, 
				  const FourMomentum<UU> & p2,
				  const FourMomentum<VV> & p3, 
				  const FourMomentum<WW> & p4,
				  const FourMomentum<XX> & p5, 
				  const FourMomentum<YY> & p6,
				  const FourMomentum<ZZ> & p7,
				  const FourMomentum<TU> & p8);

  template<typename TT, typename UU, typename VV, typename WW, typename XX,
       typename YY,typename ZZ,typename TU>
  friend std::complex<DOUBLE> IBB6(const FourMomentum<TT> & p1, 
				  const FourMomentum<UU> & p2,
				  const FourMomentum<VV> & p3, 
				  const FourMomentum<WW> & p4,
				  const FourMomentum<XX> & p5, 
				  const FourMomentum<YY> & p6,
				  const FourMomentum<ZZ> & p7,
				  const FourMomentum<TU> & p8);


};

//
//  End class definition
//

template<typename T>
void FourMomentum<T>::setFourMomentum(const FourMomentum<T> & p){
  mom[0]=p.mom[0];
  mom[1]=p.mom[1];
  mom[2]=p.mom[2];
  mom[3]=p.mom[3];
}

template<typename T>
void FourMomentum<T>::setFourMomentum(const T p[4]){
  mom[0] = p[0]; mom[1] = p[1]; mom[2] = p[2]; mom[3] = p[3];
}

template<typename T>
void FourMomentum<T>::setFourMomentum(const T p0,const T p1, 
		       const T p2, const T p3){

  mom[0] = p0; mom[1] = p1; mom[2] = p2; mom[3] = p3;

}

// private member function

template<typename TT>
const FourMomentum<TT>  operator-(const FourMomentum<TT> & p1){
  return( FourMomentum<TT>( 
		       -p1.mom[0],-p1.mom[1],-p1.mom[2],-p1.mom[3]) );
}

template<typename TT>
const FourMomentum<TT> & operator+(const FourMomentum<TT> & p1){
  return(p1);
}

template<typename TT>
const FourMomentum<TT> operator-(const FourMomentum<TT> & p1, 
			     const FourMomentum<TT> & p2){
  return(FourMomentum<TT>(
		       p1.mom[0] - p2.mom[0],
		       p1.mom[1] - p2.mom[1],
		       p1.mom[2] - p2.mom[2],
		       p1.mom[3] - p2.mom[3]
		       )
	 );
}

template<typename TT, typename CC>
  FourMomentum<std::complex<DOUBLE> > operator-(const FourMomentum<TT> & p1, 
						const FourMomentum<CC> & p2){
  return(FourMomentum<std::complex<DOUBLE> >(
					      p1.mom[0] - p2.mom[0],
					      p1.mom[1] - p2.mom[1],
					      p1.mom[2] - p2.mom[2],
					      p1.mom[3] - p2.mom[3]
					      )
	 );
}


template<typename TT>
const FourMomentum<TT> operator+(const FourMomentum<TT> & p1, 
			       const FourMomentum<TT> & p2){
  return(FourMomentum<TT>(
			   p1.mom[0] + p2.mom[0],
			   p1.mom[1] + p2.mom[1],
			   p1.mom[2] + p2.mom[2],
			   p1.mom[3] + p2.mom[3]
			   ));
}


template<typename TT, typename CC>
  FourMomentum<std::complex<DOUBLE> > operator+(const FourMomentum<TT> & p1, 
						const FourMomentum<CC> & p2){
  return(FourMomentum<std::complex<DOUBLE> >(
				       	 p1.mom[0] + p2.mom[0],
					 p1.mom[1] + p2.mom[1],
					 p1.mom[2] + p2.mom[2],
					 p1.mom[3] + p2.mom[3]
					 ));
}


template<typename TT>
const FourMomentum<TT> operator*(const DOUBLE & s, const FourMomentum<TT> & p){
  return(FourMomentum<TT>( s*p.mom[0],
			   s*p.mom[1],
			   s*p.mom[2],
			   s*p.mom[3]));
}

template<typename TT>
const FourMomentum<TT> operator*(const FourMomentum<TT> & p,const DOUBLE & s ){
  return(FourMomentum<TT>( s*p.mom[0],
			   s*p.mom[1],
			   s*p.mom[2],
			   s*p.mom[3]
			   ));
}


template<typename TT>
const FourMomentum<std::complex<DOUBLE> > operator*
( const std::complex<DOUBLE> & s, const FourMomentum<TT> & p){
  return(FourMomentum<std::complex<DOUBLE> >( s*p.mom[0],
					      s*p.mom[1],
					      s*p.mom[2],
					      s*p.mom[3]));
}

template<typename TT>
const FourMomentum<TT> operator*(const FourMomentum<std::complex<DOUBLE> > & p, 
				 const std::complex<DOUBLE> & s ){
  return(FourMomentum<std::complex<DOUBLE> >( s*p.mom[0],
					 s*p.mom[1],
					 s*p.mom[2],
					 s*p.mom[3]
					 ));
}

template<typename TT>
const FourMomentum<TT> operator/(const FourMomentum<TT> & p,const DOUBLE & s ){
  return(FourMomentum<TT>( p.mom[0]/s,
			   p.mom[1]/s,
			   p.mom[2]/s,
			   p.mom[3]/s
			   ));
}

template<typename TT>
FourMomentum<std::complex<DOUBLE> > operator/(const FourMomentum<TT> & p,
					 const std::complex<DOUBLE> & s ){
  return(FourMomentum<std::complex<DOUBLE> >( p.mom[0]/s,
					 p.mom[1]/s,
					 p.mom[2]/s,
					 p.mom[3]/s
					 ));
}

template<typename TT>
std::ostream& operator<<(std::ostream& os, const FourMomentum<TT> & P){
  return( os << "{"<< P.mom[0]<<"; " 
	  << P.mom[1] << "; "
	  << P.mom[2] << "; "
	  << P.mom[3] << "}"<<std::endl);
}

template<typename TT>
TT dot(const FourMomentum<TT> & p1, const FourMomentum<TT> & p2){
	return( p1.mom[0]*p2.mom[0]
		-p1.mom[1]*p2.mom[1]-p1.mom[2]*p2.mom[2]-p1.mom[3]*p2.mom[3]);
}


template<typename TT, typename CC>
  std::complex<DOUBLE> dot(const FourMomentum<TT> & p1, 
			   const FourMomentum<CC> & p2){
  return( p1.mom[0]*p2.mom[0]
	  -p1.mom[1]*p2.mom[1]-p1.mom[2]*p2.mom[2]-p1.mom[3]*p2.mom[3]);
}

template<typename TT>
TT S(const FourMomentum<TT> & P){
return dot(P,P);
}

template<typename TT>
TT IS(const FourMomentum<TT> & P){
	return 1/dot(P,P);
}

//
// we define the momentum matrix as:
//
//              ( p0+p2    p3-i*p1 )   ( pp(p)  pTb(p) )
//  [sigma].p = (                  ) = (               )
//              ( p3+i*p1  p0-p2   )   ( pT(p)  pm(p)  )
//
// NB.  1,2,3 -> 3,2,1 w.r.t. standard definitions 
// so incoming particles can be defined along the z-axis

template<typename TT>
std::complex<DOUBLE> pp(const FourMomentum<TT> & p){
  return p.mom[0]+p.mom[2];
}


template<typename TT>
std::complex<DOUBLE> pm(const FourMomentum<TT> & p){
  return p.mom[0]-p.mom[2];
}

template<typename TT>
std::complex<DOUBLE> pT(const FourMomentum<TT> & p){
  return (p.mom[3]+std::complex<DOUBLE>(0,DOUBLE(1.))*p.mom[1]);
}

template<typename TT>
std::complex<DOUBLE> pTb(const FourMomentum<TT> & p){
  return (p.mom[3]-std::complex<DOUBLE>(0,DOUBLE(1.))*p.mom[1]);
}

// normalisation of spinor products

template<typename TT>
std::complex<DOUBLE> SpNorm(const FourMomentum<TT> & p){
  return sqrt(abs(pp(p)))/pTb(p)*exp(std::complex<DOUBLE>(0,DOUBLE(1.))*arg(pp(p))/DOUBLE(2.));
}


//
// functions of defined as
// <ab> = A(a,b) ; [ab] = B(a,b)
// <a|b|c] = AB(a,b,c) ; [a|b|c> = BA(a,b,c)
// <a|b.c|d> = AA(a,b,c,d) ; [a|b.c|d] = BB(a,b,c,d)
// <a|b.c.d|e] = AB3(a,b,c,d,e)
// <a|b.c.d.e|f> = AA4(a,b,c,d,e,f) ; [a|b.c.d.e|f] = BB4(a,b,c,d,e,f)
// inverse functions IA() = 1/A() etc.
//
//********************************
//
//    spinor products
//
//********************************
//
//

template<typename TT, typename UU>
  std::complex<DOUBLE> A(const FourMomentum<TT> & p1,
			 const FourMomentum<UU> & p2) {
  return ( SpNorm(p1)*SpNorm(p2)*( pm(p1)*pTb(p2)-pm(p2)*pTb(p1) ) );
}

template<typename TT, typename UU>
  std::complex<DOUBLE> IA(const FourMomentum<TT> & p1,
			  const FourMomentum<UU> & p2){
  return DOUBLE(1.)/A(p1,p2);
}

template<typename TT, typename UU>
std::complex<DOUBLE> B(const FourMomentum<TT> & p1,
		       const FourMomentum<UU> & p2){
  return DOUBLE(1.)/SpNorm(p1)/SpNorm(p2)*( pp(p1)/pTb(p1)-pp(p2)/pTb(p2) );
}

template<typename TT, typename UU>
  std::complex<DOUBLE> IB(const FourMomentum<TT> & p1,
			  const FourMomentum<UU> & p2){
  return DOUBLE(1.)/B(p1,p2);
}

//{{{ automaitcally generated functions for spinor strings

template<typename T1,typename T2,typename T3>
std::complex<DOUBLE> AB(const FourMomentum<T1> &p1,const FourMomentum<T2> &p2,const FourMomentum<T3> &p3)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t9;
  {
    t1 = pp(p2);
    t2 = pTb(p2);
    t3 = pp(p3);
    t5 = pTb(p3);
    t6 = DOUBLE(1.)/t5;
    t9 = pm(p1);
    t11 = pT(p2);
    t12 = pm(p2);
    t16 = pTb(p1);
    t19 = SpNorm(p3);
    t22 = SpNorm(p1);
    return(((t1-t2*t3*t6)*t9-(t11-t12*t3*t6)*t16)/t19*t22);
  }
}

template<typename T1,typename T2,typename T3,typename T4>
std::complex<DOUBLE> AA(const FourMomentum<T1> &p1,const FourMomentum<T2> &p2,const FourMomentum<T3> &p3,const FourMomentum<T4> &p4)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t8;
  {
    t1 = pp(p2);
    t2 = pm(p3);
    t3 = pTb(p4);
    t5 = pTb(p3);
    t6 = pm(p4);
    t8 = t2*t3-t5*t6;
    t10 = pTb(p2);
    t11 = pT(p3);
    t13 = pp(p3);
    t15 = -t11*t3+t13*t6;
    t18 = pm(p1);
    t20 = pT(p2);
    t22 = pm(p2);
    t25 = pTb(p1);
    t28 = SpNorm(p4);
    t30 = SpNorm(p1);
    return(((t1*t8+t10*t15)*t18-(t20*t8+t22*t15)*t25)*t28*t30);
  }
}


template<typename T1,typename T2,typename T3,typename T4>
std::complex<DOUBLE> BB(const FourMomentum<T1> &p1,const FourMomentum<T2> &p2,const FourMomentum<T3> &p3,const FourMomentum<T4> &p4)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t9;
  {
    t1 = pm(p2);
    t2 = pp(p3);
    t3 = pTb(p3);
    t4 = pp(p4);
    t6 = pTb(p4);
    t7 = DOUBLE(1.)/t6;
    t9 = t2-t3*t4*t7;
    t11 = pTb(p2);
    t12 = pT(p3);
    t13 = pm(p3);
    t16 = t12-t13*t4*t7;
    t19 = pp(p1);
    t21 = pTb(p1);
    t24 = pT(p2);
    t26 = pp(p2);
    t29 = SpNorm(p4);
    t32 = SpNorm(p1);
    return(((t1*t9-t11*t16)*t19/t21-t24*t9+t26*t16)/t29/t32);
  }
}


template<typename T1,typename T2,typename T3,typename T4,typename T5>
std::complex<DOUBLE> AB3(const FourMomentum<T1> &p1,const FourMomentum<T2> &p2,const FourMomentum<T3> &p3,const FourMomentum<T4> &p4,const FourMomentum<T5> &p5)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t8;
  {
    t1 = pp(p2);
    t2 = pm(p3);
    t3 = pp(p4);
    t4 = pTb(p4);
    t5 = pp(p5);
    t7 = pTb(p5);
    t8 = DOUBLE(1.)/t7;
    t10 = t3-t4*t5*t8;
    t12 = pTb(p3);
    t13 = pT(p4);
    t14 = pm(p4);
    t17 = t13-t14*t5*t8;
    t19 = t2*t10-t12*t17;
    t21 = pTb(p2);
    t22 = pT(p3);
    t24 = pp(p3);
    t26 = -t22*t10+t24*t17;
    t29 = pm(p1);
    t31 = pT(p2);
    t33 = pm(p2);
    t36 = pTb(p1);
    t39 = SpNorm(p5);
    t42 = SpNorm(p1);
    return(((t1*t19+t21*t26)*t29-(t31*t19+t33*t26)*t36)/t39*t42);
  }
}


template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6>
std::complex<DOUBLE> AA4(const FourMomentum<T1> &p1,const FourMomentum<T2> &p2,const FourMomentum<T3> &p3,const FourMomentum<T4> &p4,const FourMomentum<T5> &p5,const FourMomentum<T6> &p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t8;
  {
    t1 = pp(p2);
    t2 = pm(p3);
    t3 = pp(p4);
    t4 = pm(p5);
    t5 = pTb(p6);
    t7 = pTb(p5);
    t8 = pm(p6);
    t10 = t4*t5-t7*t8;
    t12 = pTb(p4);
    t13 = pT(p5);
    t15 = pp(p5);
    t17 = -t13*t5+t15*t8;
    t19 = t3*t10+t12*t17;
    t21 = pTb(p3);
    t22 = pT(p4);
    t24 = pm(p4);
    t26 = t22*t10+t24*t17;
    t28 = t2*t19-t21*t26;
    t30 = pTb(p2);
    t31 = pT(p3);
    t33 = pp(p3);
    t35 = -t31*t19+t33*t26;
    t38 = pm(p1);
    t40 = pT(p2);
    t42 = pm(p2);
    t45 = pTb(p1);
    t48 = SpNorm(p6);
    t50 = SpNorm(p1);
    return(((t1*t28+t30*t35)*t38-(t40*t28+t42*t35)*t45)*t48*t50);
  }
}


template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6>
std::complex<DOUBLE> BB4(const FourMomentum<T1> &p1,const FourMomentum<T2> &p2,const FourMomentum<T3> &p3,const FourMomentum<T4> &p4,const FourMomentum<T5> &p5,const FourMomentum<T6> &p6)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t18;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t8;
  std::complex<DOUBLE> t9;
  {
    t1 = pm(p2);
    t2 = pp(p3);
    t3 = pm(p4);
    t4 = pp(p5);
    t5 = pTb(p5);
    t6 = pp(p6);
    t8 = pTb(p6);
    t9 = DOUBLE(1.)/t8;
    t11 = t4-t5*t6*t9;
    t13 = pTb(p4);
    t14 = pT(p5);
    t15 = pm(p5);
    t18 = t14-t15*t6*t9;
    t20 = t11*t3-t13*t18;
    t22 = pTb(p3);
    t23 = pT(p4);
    t25 = pp(p4);
    t27 = -t23*t11+t25*t18;
    t29 = t2*t20+t22*t27;
    t31 = pTb(p2);
    t32 = pT(p3);
    t34 = pm(p3);
    t36 = t32*t20+t34*t27;
    t39 = pp(p1);
    t41 = pTb(p1);
    t44 = pT(p2);
    t46 = pp(p2);
    t49 = SpNorm(p6);
    t52 = SpNorm(p1);
    return(((t1*t29-t31*t36)*t39/t41-t44*t29+t46*t36)/t49/t52);
  }
}


template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7>
std::complex<DOUBLE> AB5(const FourMomentum<T1> &p1,const FourMomentum<T2> &p2,const FourMomentum<T3> &p3,const FourMomentum<T4> &p4,const FourMomentum<T5> &p5,const FourMomentum<T6> &p6,
const FourMomentum<T7> &p7)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t9;
  {
    t1 = pp(p2);
    t2 = pm(p3);
    t3 = pp(p4);
    t4 = pm(p5);
    t5 = pp(p6);
    t6 = pTb(p6);
    t7 = pp(p7);
    t9 = pTb(p7);
    t10 = DOUBLE(1.)/t9;
    t12 = t5-t6*t7*t10;
    t14 = pTb(p5);
    t15 = pT(p6);
    t16 = pm(p6);
    t19 = t15-t16*t7*t10;
    t21 = t4*t12-t14*t19;
    t23 = pTb(p4);
    t24 = pT(p5);
    t26 = pp(p5);
    t28 = -t24*t12+t26*t19;
    t30 = t3*t21+t23*t28;
    t32 = pTb(p3);
    t33 = pT(p4);
    t35 = pm(p4);
    t37 = t33*t21+t35*t28;
    t39 = t2*t30-t32*t37;
    t41 = pTb(p2);
    t42 = pT(p3);
    t44 = pp(p3);
    t46 = -t42*t30+t44*t37;
    t49 = pm(p1);
    t51 = pT(p2);
    t53 = pm(p2);
    t56 = pTb(p1);
    t59 = SpNorm(p7);
    t62 = SpNorm(p1);
    return(((t1*t39+t41*t46)*t49-(t51*t39+t53*t46)*t56)/t59*t62);
  }
}


template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8>
std::complex<DOUBLE> AA6(const FourMomentum<T1> &p1,const FourMomentum<T2> &p2,const FourMomentum<T3> &p3,const FourMomentum<T4> &p4,const FourMomentum<T5> &p5,const FourMomentum<T6> &p6,
const FourMomentum<T7> &p7,const FourMomentum<T8> &p8)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t12;
  std::complex<DOUBLE> t14;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t19;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t21;
  std::complex<DOUBLE> t23;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t26;
  std::complex<DOUBLE> t28;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t30;
  std::complex<DOUBLE> t32;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t35;
  std::complex<DOUBLE> t37;
  std::complex<DOUBLE> t39;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t41;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t44;
  std::complex<DOUBLE> t46;
  std::complex<DOUBLE> t48;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t50;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t53;
  std::complex<DOUBLE> t55;
  std::complex<DOUBLE> t58;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t60;
  std::complex<DOUBLE> t62;
  std::complex<DOUBLE> t65;
  std::complex<DOUBLE> t68;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t70;
  std::complex<DOUBLE> t9;
  {
    t1 = pp(p2);
    t2 = pm(p3);
    t3 = pp(p4);
    t4 = pm(p5);
    t5 = pp(p6);
    t6 = pm(p7);
    t7 = pTb(p8);
    t9 = pTb(p7);
    t10 = pm(p8);
    t12 = t6*t7-t9*t10;
    t14 = pTb(p6);
    t15 = pT(p7);
    t17 = pp(p7);
    t19 = -t15*t7+t17*t10;
    t21 = t5*t12+t14*t19;
    t23 = pTb(p5);
    t24 = pT(p6);
    t26 = pm(p6);
    t28 = t24*t12+t26*t19;
    t30 = t4*t21-t23*t28;
    t32 = pTb(p4);
    t33 = pT(p5);
    t35 = pp(p5);
    t37 = -t33*t21+t35*t28;
    t39 = t3*t30+t32*t37;
    t41 = pTb(p3);
    t42 = pT(p4);
    t44 = pm(p4);
    t46 = t42*t30+t44*t37;
    t48 = t2*t39-t41*t46;
    t50 = pTb(p2);
    t51 = pT(p3);
    t53 = pp(p3);
    t55 = -t51*t39+t53*t46;
    t58 = pm(p1);
    t60 = pT(p2);
    t62 = pm(p2);
    t65 = pTb(p1);
    t68 = SpNorm(p8);
    t70 = SpNorm(p1);
    return(((t1*t48+t50*t55)*t58-(t60*t48+t62*t55)*t65)*t68*t70);
  }
}

template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8>
std::complex<DOUBLE> BB6(const FourMomentum<T1> &p1,const FourMomentum<T2> &p2,const FourMomentum<T3> &p3,const FourMomentum<T4> &p4,const FourMomentum<T5> &p5,const FourMomentum<T6> &p6,
const FourMomentum<T7> &p7,const FourMomentum<T8> &p8)
{
  std::complex<DOUBLE> t1;
  std::complex<DOUBLE> t10;
  std::complex<DOUBLE> t11;
  std::complex<DOUBLE> t13;
  std::complex<DOUBLE> t15;
  std::complex<DOUBLE> t16;
  std::complex<DOUBLE> t17;
  std::complex<DOUBLE> t2;
  std::complex<DOUBLE> t20;
  std::complex<DOUBLE> t22;
  std::complex<DOUBLE> t24;
  std::complex<DOUBLE> t25;
  std::complex<DOUBLE> t27;
  std::complex<DOUBLE> t29;
  std::complex<DOUBLE> t3;
  std::complex<DOUBLE> t31;
  std::complex<DOUBLE> t33;
  std::complex<DOUBLE> t34;
  std::complex<DOUBLE> t36;
  std::complex<DOUBLE> t38;
  std::complex<DOUBLE> t4;
  std::complex<DOUBLE> t40;
  std::complex<DOUBLE> t42;
  std::complex<DOUBLE> t43;
  std::complex<DOUBLE> t45;
  std::complex<DOUBLE> t47;
  std::complex<DOUBLE> t49;
  std::complex<DOUBLE> t5;
  std::complex<DOUBLE> t51;
  std::complex<DOUBLE> t52;
  std::complex<DOUBLE> t54;
  std::complex<DOUBLE> t56;
  std::complex<DOUBLE> t59;
  std::complex<DOUBLE> t6;
  std::complex<DOUBLE> t61;
  std::complex<DOUBLE> t64;
  std::complex<DOUBLE> t66;
  std::complex<DOUBLE> t69;
  std::complex<DOUBLE> t7;
  std::complex<DOUBLE> t72;
  std::complex<DOUBLE> t8;
  {
    t1 = pm(p2);
    t2 = pp(p3);
    t3 = pm(p4);
    t4 = pp(p5);
    t5 = pm(p6);
    t6 = pp(p7);
    t7 = pTb(p7);
    t8 = pp(p8);
    t10 = pTb(p8);
    t11 = DOUBLE(1.)/t10;
    t13 = t6-t7*t8*t11;
    t15 = pTb(p6);
    t16 = pT(p7);
    t17 = pm(p7);
    t20 = t16-t17*t8*t11;
    t22 = t13*t5-t15*t20;
    t24 = pTb(p5);
    t25 = pT(p6);
    t27 = pp(p6);
    t29 = -t25*t13+t27*t20;
    t31 = t4*t22+t24*t29;
    t33 = pTb(p4);
    t34 = pT(p5);
    t36 = pm(p5);
    t38 = t34*t22+t36*t29;
    t40 = t3*t31-t33*t38;
    t42 = pTb(p3);
    t43 = pT(p4);
    t45 = pp(p4);
    t47 = -t43*t31+t45*t38;
    t49 = t2*t40+t42*t47;
    t51 = pTb(p2);
    t52 = pT(p3);
    t54 = pm(p3);
    t56 = t52*t40+t54*t47;
    t59 = pp(p1);
    t61 = pTb(p1);
    t64 = pT(p2);
    t66 = pp(p2);
    t69 = SpNorm(p8);
    t72 = SpNorm(p1);
    return(((t1*t49-t51*t56)*t59/t61-t64*t49+t66*t56)/t69/t72);
  }
}

//}}}

template<typename TT, typename UU,typename VV>
  std::complex<DOUBLE> IAB(const FourMomentum<TT> & p1, 
			   const FourMomentum<UU> & p2,
			   const FourMomentum<VV> & p3){
  return DOUBLE(1.)/AB(p1,p2,p3);
}

template<typename TT, typename UU, typename VV>
  std::complex<DOUBLE> BA(const FourMomentum<TT> & p1, 
			  const FourMomentum<UU> & p2,
			  const FourMomentum<VV> & p3){
  return AB(p3,p2,p1);
}

template<typename TT, typename UU, typename VV>
  std::complex<DOUBLE> IBA(const FourMomentum<TT> & p1, 
			   const FourMomentum<UU> & p2,
			   const FourMomentum<VV> & p3){
  return DOUBLE(1.)/AB(p3,p2,p1);
}

template<typename TT, typename UU, typename VV, typename WW>
  std::complex<DOUBLE> IAA(const FourMomentum<TT> & p1, 
			   const FourMomentum<UU> & p2,
			   const FourMomentum<VV> & p3,
			   const FourMomentum<WW> & p4){
  return DOUBLE(1.)/AA(p1,p2,p3,p4);
}

template<typename TT, typename UU, typename VV, typename WW>
  std::complex<DOUBLE> IBB(const FourMomentum<TT> & p1, 
			   const FourMomentum<UU> & p2,
			   const FourMomentum<VV> & p3, 
			   const FourMomentum<WW> & p4){
  return DOUBLE(1.)/BB(p1,p2,p3,p4);
}

template<typename TT, typename UU, 
  typename VV, typename WW, typename XX>
  std::complex<DOUBLE> IAB3(const FourMomentum<TT> & p1, 
			    const FourMomentum<UU> & p2,
			    const FourMomentum<VV> & p3, 
			    const FourMomentum<WW> & p4,
			    const FourMomentum<XX> & p5){
  return DOUBLE(1.)/AB3(p1,p2,p3,p4,p5);
}

template<typename TT, typename UU, 
  typename VV, typename WW, typename XX>
  std::complex<DOUBLE> BA3(const FourMomentum<TT> & p1, 
			    const FourMomentum<UU> & p2,
			    const FourMomentum<VV> & p3, 
			    const FourMomentum<WW> & p4,
			    const FourMomentum<XX> & p5){
  return AB3(p5,p4,p3,p2,p1);
}


template<typename TT, typename UU, typename VV, 
  typename WW, typename XX, typename YY>
  std::complex<DOUBLE> IAA4(const FourMomentum<TT> & p1, 
			    const FourMomentum<UU> & p2,
			    const FourMomentum<VV> & p3, 
			    const FourMomentum<WW> & p4,
			    const FourMomentum<XX> & p5, 
			    const FourMomentum<YY> & p6){
  return DOUBLE(1.)/AA4(p1,p2,p3,p4,p5,p6);
}

template<typename TT, typename UU, typename VV, 
  typename WW, typename XX, typename YY>
  std::complex<DOUBLE> IBB4(const FourMomentum<TT> & p1, 
			    const FourMomentum<UU> & p2,
			    const FourMomentum<VV> & p3, 
			    const FourMomentum<WW> & p4,
			    const FourMomentum<XX> & p5,
			    const FourMomentum<YY> & p6){
  return DOUBLE(1.)/BB4(p1,p2,p3,p4,p5,p6);
}

#endif // _FOURMOMENTUM_H
