#ifndef PHASESPACE_H_
#define PHASESPACE_H_

#include <iostream>
#include <fstream>
#include "mytypes.h"
#include "FourMomentum.h"

int eventn(const unsigned int n, const DOUBLE sqrts, const double x[], 
           const DOUBLE masses[], DOUBLE kout[][4], DOUBLE& jacobi);
struct RandNums {

  static const int seed = 295;
  double* xx;
  const int len;

  ~RandNums(){
    delete [] xx;
  };

  RandNums(const int len_) : len(len_) {
    xx = new double[3*(len-2)-4]; 
    // intialise random numbers
    srand48(seed);
  };

  void getRandomNumbers(){
    for(int i = 0; i < 3*(len-2)-4; i++){
      xx[i] = floor(drand48()*10e8)/10e8;
    }
  }

  void printRandomNumbers(){
    std::cout << "random numbers for phase-space generation:" << std::endl;
    std::cout.precision(12);
    for(int i = 0; i < 3*(len-2)-4; i++){
      std::cout << xx[i] << "  "; 
    }
    std::cout << std::endl;
  }

  void printRandomNumbers(std::fstream& file){
    file << "random numbers for phase-space generation:" << std::endl;
    file.precision(12);
    file << "{";
    for(int i = 0; i < 3*(len-2)-4; i++){
      file << xx[i] << ",  "; 
    }
   file << "}" << std::endl;
  }

  void setRandomNumbers(double xxx[]){
    for(int i = 0; i < 3*(len-2)-4; i++){
      xx[i] = xxx[i];
    }
  }
};


class KinCuts {
 public:
  
 KinCuts(DOUBLE (*k_)[4], int nn_, DOUBLE cmEnergy_, std::string cut_) 
   :  cut(cut_+"-cuts"), nn(nn_), cmEnergy(cmEnergy_), moms(k_)  {

    pT  = new DOUBLE [nn];
    eta = new DOUBLE [nn];
    Phi = new DOUBLE [nn];

    count = 0;

    if( cut_ == "GKM" ){ // Giele, Kunszt and Melnikov
      info("kinematical cuts like in [arXiv:0708.2398] used");
      set_etaMax(3.0);
      set_pTMin(cmEnergy * 0.01);
      set_RijMin(0.4);
      function = &KinCuts::etaPhiPT;
      return;
    }
    
    if( cut_ == "GKW" ){ // Giele, Kunszt and Winter
      info("kinematical cuts like in [arXiv:0911.1962] used");
      set_etaMax(2.0);
      set_pTMin(cmEnergy * 0.1);
      set_RijMin(0.7);
      function = &KinCuts::etaPhiPT;
      return;
    }
    
    if( cut_ == "jade" ){
      info("kinematical cuts: jade");
      set_deltaS(0.001);
      function = &KinCuts::jade;
      return;
    }

    info("no kinematic cuts specified");
    function = &KinCuts::noCuts;

  };
  
  ~KinCuts(){
    delete [] pT;
    delete [] eta;
    delete [] Phi;
  };

  int count;
  const std::string cut;
  
  void set_cmEnergy(const DOUBLE cmEnergy_);
  void set_pTMin(const DOUBLE pTMin_);
  void set_etaMax(const DOUBLE etaMax_);
  void set_RijMin(const DOUBLE RijMin_);
  void set_deltaS(const DOUBLE deltaS_);
  
  bool check();
  bool etaPhiPT();
  bool jade();
  bool noCuts(){ return (0);}
  
  DOUBLE rapidity(DOUBLE p[4]);
  DOUBLE transvMom(DOUBLE p[4]);
  DOUBLE phi(DOUBLE p[4]);
  DOUBLE dot(DOUBLE p[4], DOUBLE q[4]);
  
 private:
  const int nn;
  DOUBLE cmEnergy, etaMax, pTMin, RijMin, deltaR, deltaS;
  DOUBLE *pT, *eta, *Phi;
  DOUBLE (*moms)[4];
  bool (KinCuts::*function)();
  void info(std::string s);
  void info(std::string s, DOUBLE tmp);
};


void momGen(const int ngluon, const DOUBLE cmEnergy, DOUBLE kk[][4],
	    RandNums& rn);
void mom2FourMom(const int ngluon, DOUBLE kk[][4], FourMomentum<DOUBLE> KK[]);
void momGen(FourMomentum<DOUBLE> *KK, const int ngluon,
	    const DOUBLE cmEnergy, RandNums& rn);

#endif // PHASESPACE_H_
